/*
 * This is a small little hack driver to allow userspace to allocate
 * physically contiguous memory, for use with DMA.
 * It is assumed to be used in conjunction with an X server that has
 * need of such a thing.
 *
 * As such, it is built to only allow a SINGLE PROCESS to open the device.
 * It makes the code a lot easier for me to deal with.
 * I can hardcode stuff, instead of being fancy ;-)
 *
 * BTW, it was built on my PCIbase driver skeleton from 
 *   from http://www.bolthole.com/solaris/drivers/
 *
 * REMEMBER: If a C function isn't mentioned in
 *  "man -s9f intro", you pretty much cannot use it in here.
 *
 * 
 */


/* Define USE_MMAP_ONLY, if Phil hasnt found a way to make
 * devmap_umem_setup() happy yet  [devmap() needs a valid umem_cookie]
 * If this is set, we only use the old mmap() entrypoint instead of
 * the newfangled spiffy devmap() entrypoint for user mmaping attempts.
 * devmap is MORE RELIABLE. mmap is FLAKEY.
 * But unfortunately, I have no choice at this point.
 */

#define USE_MMAP_ONLY


#include <varargs.h>
#define _SYS_VARARGS_H		/* sys/varargs.h hack for gcc */

#include <sys/errno.h>
#include <sys/conf.h>
#include <sys/modctl.h>
#include <sys/open.h>
#include <sys/stat.h>

#include "dmamem.h"

/* THESE MUST BE LAST of the system includes */
#include <sys/ddi.h>
#include <sys/sunddi.h>



/* This ASSUMES you create minor nodes where the higher bits denote
 * the instance of the driver. (aka "which card is this?")
 * The lower bits denote clone number, or anything else you want.
 */

#define DEVT_TO_INST(devt) (getminor(devt))

/* Used to choose a node  num for a device node compatible with DEVT_TO_INST*/
#define INST_TO_NODENUM(instance) (instance)

/* As well as changing the NAME of this, you will also want to change
 * the TYPE of the minor node created.
 */
#define DEVNODE_NAME	"dmamem"

/* be careful what you set the type to. setting it to known DDI_NT_XXX
 * types may have side effects you may not be aware of, for devlink
 * purposes
 */
#define DEVNODE_TYPE	DDI_PSEUDO /* see sunddi.h */

/*Our globally visible stuffs, that we populate the global
 * driver entry structs with  [dmamem_cb_ops, dmamem_ops]
 * Yeah, its kind of wierd to have a "static" be also externally visible,
 * but that's traditionally how these are declared.
 */
static	int	dmamem_getinfo(dev_info_t *dip, ddi_info_cmd_t infocmd, void *arg,
		    void **result);

static	int	dmamem_attach(dev_info_t *dip, ddi_attach_cmd_t cmd);
static	int	dmamem_open(dev_t *dev, int openflags, int otyp, cred_t *credp);

static	int	dmamem_print(dev_t, char*);

static	int	dmamem_read(dev_t dev, struct uio* uiop, cred_t * credp);

static	int	dmamem_close(dev_t dev, int openflags, int otyp, cred_t *credp);
static	int	dmamem_write(dev_t dev, struct uio *uiop, cred_t *credp);
static	int	dmamem_ioctl(dev_t dev, int cmd, intptr_t arg, int flags,
			   cred_t *credp, int *rvalp);
static	int	dmamem_detach(dev_info_t *dip, ddi_detach_cmd_t cmd);

#ifdef USE_MMAP_ONLY
static	int	dmamem_mmap(dev_t dev, off_t off, int prot);
#else
static	int	dmamem_devmap(dev_t dev, devmap_cookie_t, offset_t,
                         size_t, size_t*, uint_t);
#endif

static int debug_level;

/* "char/block operations" */
static struct cb_ops	dmamem_cb_ops = {
	dmamem_open,    /* open() */
	dmamem_close,   /* close() */
	nodev,          /* strategy() */
	dmamem_print,   /* print routine */
	nodev,          /* no dump routine */
	dmamem_read,    /* read() */
	dmamem_write,   /* write() */
	dmamem_ioctl,   /* generic ioctl */
#ifdef USE_MMAP_ONLY
	nodev,          /* no devmap routine */
	dmamem_mmap,    /* mmap() handler */
#else
	dmamem_devmap,  /* devmap routine */
	nodev,          /* no mmap routine */
#endif
	nodev,          /* no segmap routine */
	nochpoll,       /* no chpoll routine */
	ddi_prop_op,
	0,              /* not a STREAMS driver	*/
#ifdef USE_MMAP_ONLY
	D_NEW | D_MP,   /* safe for multi-thread/multi-processor */
#else
	D_NEW | D_MP | D_DEVMAP,   /* safe for multi-thread/multi-processor */
#endif
	CB_REV,         /* cb_ops version? */
	nodev,          /* cb_aread() */
	nodev,          /* cb_awrite() */
};

/* "device operations". */
static struct dev_ops dmamem_ops = {
	DEVO_REV,            /* devo_rev */
	0,                   /* devo_refcnt */
	dmamem_getinfo,        /* devo_getinfo */
	nulldev,             /* devo_identify */
	nulldev,			/* devo_probe */
	dmamem_attach,         /* devo_attach */
	dmamem_detach,         /* devo_detach */
	nodev,               /* devo_reset */
	&dmamem_cb_ops,        /* devo_cb_ops */
	(struct bus_ops *)0, /* devo_bus_ops */
	NULL,                /* devo_power */
};

/* Ignore this. This is only a filler for the next struct, and
 * is already declared in <sys/modctl.h>
 */
extern	struct	mod_ops mod_driverops;

static	struct modldrv modldrv = {
	&mod_driverops,
	"Experimental _dmamem_ driver",
	&dmamem_ops,
};

static	struct modlinkage modlinkage = {
	MODREV_1,		/* MODREV_1 is indicated by manual */
	{&modldrv,NULL,NULL,NULL}

};


/************************************************************
 * And here's finally the vars specific to THIS driver,
 * that get allocated on a per-instance basis.
 * Here's where you should keep instance-specific mutexes,
 * DMA handles, register mapping handles, and stuff like that
 *
 ************************************************************/


/* This describes things like byteorder internally for the pci card */
static struct ddi_device_acc_attr dev_attr = {
	DDI_DEVICE_ATTR_V0,
	DDI_STRUCTURE_LE_ACC,
	DDI_STRICTORDER_ACC
};



#define MAX_INSTANCES 1 /* max of 1 driver loaded */

/* 
 * Note that this mutex gets used in _init. Which means this mutex
 * isn't just for this "instance" of the driver. It is for ALL INSTANCES.
 * If you have multiple cards, you have multiple instances of the driver,
 * with this single shared mutex across them all
 */
static kmutex_t   dmamem_global_mutex;

/* This is SORT OF like	int (*array)[] , except you NEVER USE THIS DIRECTLY.
 * You just use it with ddi_get_soft_state().
 * 
 *    "Global Softstate Handle" 
 */
void * dmamem_glob_soft_handle;


/* Holds all the relevant information about a single allocated memchunk */
/* If memlen ==0, that means this struct is still uninitialized, aka unused*/
struct memchunk {
	ddi_dma_handle_t dmahandle;
	uint32_t memlen;
	size_t reallen;
	caddr_t kaddr;
	ddi_acc_handle_t dma_acc_handle; /* "access" handle */

	/* These two are used if we use ddi_dma_addr_bind_handle() */
	ddi_dma_cookie_t dmacookie;
	uint_t ccount;	/* cookie count */

	/* This is used if we use ddi_umem_lock() */
	ddi_umem_cookie_t umem_cookie;

	uint32_t offset; /* offset into our own driver-virtual memory space*/
};


/* This gets filled out on a per-instance basis by dmamem_attach */
struct dmamem_softstate {
	dev_info_t *dip; /* you'll probably want this saved! */
	kmutex_t instance_mutex;

	struct memchunk chunks[DMAMEM_MAX_CHUNKS];

	int openflag;

};



/*****************************************************************/
/*****************************************************************/
/*              Start of actual kernel routines here             */
/*****************************************************************/
/*****************************************************************/



void
dmamem_debugprint(va_alist)
va_dcl
{
	va_list args;
	char *fmt;
	if(debug_level==0){
		return;
	}

	va_start(args);
	fmt = va_arg(args, char *);
	/* DO NOT USE cmn_err here!!! it will not be interrupt-safe */
	vprintf(fmt, args);
	va_end(args);
}



/*Like the dmamem_global_mutex above, this routine is GLOBAL across all
 * instances.
 * It actually gets called when you add_drv your driver from the command
 * line, before  * it is hooked up to any specific hardware.
 * See dmamem_attach() below for instance-specific stuff.
 *
 * HOWEVER... Its often nice to pre-allocate "softstate" space
 * here, if you dont mind hardcoding the total number of possible
 * instances you can have. So we do that.
 * 
 */
int
_init(void)
{
       int     ret;

	dmamem_debugprint("dmamem: _init called\n");
	/*
	 * Initialize global mutex before mod_install'ing driver.
	 * If mod_install() fails, must clean up mutex initialization
	 */
	mutex_init(&dmamem_global_mutex, "dmamem: Global Mutex",
                     MUTEX_DRIVER, (void *)NULL);

	/* This is kind of like
	 * malloc(sizeof(struct dmamem_softstate *MAX_INSTANCES))
	 */
	ret=ddi_soft_state_init(&dmamem_glob_soft_handle,
				sizeof(struct dmamem_softstate),
			        MAX_INSTANCES);
	if(ret!=0){
		return (ret);
	}
	
	if ((ret = mod_install(&modlinkage)) != 0) {
		mutex_destroy(&dmamem_global_mutex);
		ddi_soft_state_fini(&dmamem_glob_soft_handle);
		cmn_err(CE_NOTE, "dmamem: _init: mod_install FAILED:\n");
		return DDI_FAILURE;
	}

	cmn_err(CE_NOTE, "dmamem: _init returning SUCCESS, supposedly\n");
	return (DDI_SUCCESS);
}

int
_info(struct modinfo *modinfop)
{
	
	cmn_err(CE_NOTE, "dmamem: _info called\n");
	return (mod_info(&modlinkage, modinfop));
}

/* This is the match to the _init() routine. Clean up stuff allocated
 * specifically in _init() here.
 * Stuff in attach() should be cleaned up in detach()
 */
int
_fini(void)
{
	int ret;

	cmn_err(CE_NOTE, "dmamem: _fini called\n");
	
        /*
         * If mod_remove() is successful, we destroy our global mutex
         */
	if ((ret = mod_remove(&modlinkage)) == 0) {
		ddi_soft_state_fini(&dmamem_glob_soft_handle);
		mutex_destroy(&dmamem_global_mutex);
	}
	return (ret);
}

/************************************************************/


/* Kinda wierd, but this is an official solaris entry point, that can
 * be registered by cb_ops. We dont use it, but we provide it
 * for the OS
 */
static int dmamem_print(dev_t dev, char*string){
	int instance=DEVT_TO_INST(dev);
	cmn_err(CE_NOTE, "base-print(%d): %s",instance,string);
	return 0;
}

static int dmamem_getinfo(dev_info_t *dip, ddi_info_cmd_t infocmd,
			     void *arg, void **result){
	dmamem_debugprint("dmamem: getinfo called\n");
	return DDI_FAILURE;
}



/* free up the memory referenced by the given memchunk, and mark
 * the memchunk as no longer used.
 * Does not deallocate the struct memchunk itself
 * (we normally dont need to do that)
 */
static void free_memchunk(struct memchunk *memchunk)
{
	if(memchunk==NULL){
		dmamem_debugprint("ERR: free_memchunk passed NULL\n");
		return;
	}
	if(memchunk->memlen==0){
		return;
	}
	printf("memlen=%d; freeing physmemory@0x%x\n",
	       memchunk->memlen, memchunk->dmacookie.dmac_address);

	if(memchunk->dma_acc_handle !=NULL){
		ddi_dma_mem_free(&memchunk->dma_acc_handle);
	} else {
		printf("free_memchunk: ERR: memlen=%d, but acc handle NULL\n",
		       memchunk->memlen);
	}
	if(memchunk->dmahandle !=NULL){
		ddi_dma_unbind_handle(memchunk->dmahandle);
		ddi_dma_free_handle(&memchunk->dmahandle);
		memchunk->dmahandle=NULL;
	} else {
		printf("free_memchunk: ERR: memlen=%d, but dmahandle NULL\n",
		       memchunk->memlen);
	}
	memchunk->memlen=0;

}


/* finds a free slot to store a memory mappable memory chunk.
 * returns NULL if cant find one.
 */
static struct memchunk *find_free_memslot(struct dmamem_softstate *softstate)
{
	int mcount;

	for(mcount=0;mcount<DMAMEM_MAX_CHUNKS;mcount++){
		if(softstate->chunks[mcount].memlen==0){
			return &softstate->chunks[mcount];
		}
	}
	dmamem_debugprint("out of free memchunk slots\n");
	return NULL;

}

/* User has said, "I want to map memory at this specific offset".
 * Find the chunk related to it.
 * This actually can only be used by devmap, not mmap,
 * since mmap(9e) gets things on a page-by-page basis,
 * not just the first  memory addr of the user request.
 *
 * Note that we assume we have allocated the mem in 32bit memory only
 */
static struct memchunk *
find_chunk_by_offset(struct dmamem_softstate *softstate, uint32_t offset)
{
	struct memchunk *memchunk;
	int mcount;

	for(mcount=0;mcount<DMAMEM_MAX_CHUNKS;mcount++){
		memchunk=&softstate->chunks[mcount];
		if(memchunk->offset == offset){
			return memchunk;
		}
	}
	return NULL;
}


/* called by mmap(9e)
 * we dont get a pure beginning-of-chunk offset.
 * we get handed an offset ANYWHERE IN OUR MEMORY.
 * up to us to figure out which chunk, if any, is involved.
 */
static struct memchunk *
find_chunk_hard_way(struct dmamem_softstate *softstate, uint32_t offset)
{
	struct memchunk *memchunk;
	int mcount;

	for(mcount=0;mcount<DMAMEM_MAX_CHUNKS;mcount++){
		memchunk=&softstate->chunks[mcount];
		if(memchunk->offset <= offset){
			if(memchunk->offset + memchunk->memlen >offset){
				return memchunk;
			}
		}
	}

	return NULL;

}

/* Passed in both softstate and memchunk.
 * Determine the highest unused "offset" in our own personal driver-local
 * mmap address space, and set it for the new memchunk passed in .
 * We make the assumption that none of the existing memchunks overlap
 * We also make the assumption that if the memchunk is part of the
 * existing softstate array,that its memlen is STILL ZERO.
 * You must call this routine before setting memlen for the chunk.
 */
static void set_free_offset(struct dmamem_softstate *softstate,
                            struct memchunk *newchunk)
{
	uint32_t highestoffset=0;
	int mcount;
	struct memchunk *tmpchunk, *highestchunk=NULL;

	for(mcount=0;mcount<DMAMEM_MAX_CHUNKS;mcount++){
		tmpchunk=&softstate->chunks[mcount];
		if(debug_level>1){
		    printf("DEBUG: set_free_offset: offset %x, len %x\n",
		       tmpchunk->offset, tmpchunk->memlen);
		}

		if(tmpchunk->memlen==0){
			continue;
		}

		if(highestchunk==NULL){
			if(debug_level>1){
			    printf("DEBUG: first chunk found: offset %x, len %x\n",
			       tmpchunk->offset, tmpchunk->memlen);
			}
			highestchunk=tmpchunk;
			continue;
		}
		
		if(tmpchunk->offset <highestchunk->offset){
			continue;
		}
		if(debug_level>1){
		    printf("DEBUG: new highest chunk: offset %x, len %x\n",
		       tmpchunk->offset, tmpchunk->memlen);
		}

		highestchunk=tmpchunk;
	}

	if(highestchunk==NULL){
		highestoffset=0;
	} else {
		highestoffset=highestchunk->offset + highestchunk->memlen;
		/* make sure rounded to 4k page */
		if(highestoffset& 0xfff){
			highestoffset= (highestoffset&0xfffff000) + 0x1000;
		}
	}

	newchunk->offset=highestoffset;

	dmamem_debugprint("Setting new memchunk to offset %x\n",highestoffset);
}

/* goes through all our memchunk slots, and frees any memory we
 * may have allocated there.
 */
static void dmamem_freemem(struct dmamem_softstate *softstate){
	struct memchunk *memchunk;
	int mcount;

	for(mcount=0;mcount<DMAMEM_MAX_CHUNKS;mcount++){
		memchunk=&softstate->chunks[mcount];
		if(memchunk->memlen!=0){
			free_memchunk(memchunk);
		}
	}
	
}


/* Passed in a dmamem struct, which is a COPY of the user's struct.
 * Whatever we set, will be copied out to the user.
 * The only value we expect to be set coming in, is dstruct->memsize.
 *
 * Return suitable value for dmamem_ioctl() to pass through as its
 *   return value to the user. 
 */
static int
create_dmamem(struct dmamem_softstate *softstate, struct dmamem*dstruct)
{
	int memsize=dstruct->memsize;
	int tmp_state=DDI_FAILURE;
	struct memchunk *memchunk;

	ddi_dma_attr_t tmp_dma_attr = {
		DMA_ATTR_V0,
		0,		/* addr_low */
		0xffffffff,	/* addr_high */
		0x00ffffff,	/* DMA counter register */
		4,		/* DMA address alignment */
		0x0f,		/* DMA burstsizes */
		1,		/* minxfer effective DMA size */
		0x00ffffff,	/* max DMA xfer size */
		0x00ffffff,	/* segment boundary*/
		1,		/* s/g length */
				/*   (we want physically contigous mem) */
		1,		/* granularity of device */
				/*   (sbpci dma operates on longwords) */
		0		/* flags */
	};

	if(memsize==0){
		dmamem_debugprint("create_mem passed ZERO memsize\n");
		return EINVAL;
	}
	memchunk=find_free_memslot(softstate);
	if(memchunk==NULL){
		return ENOMEM;
	}

	tmp_state=ddi_dma_alloc_handle(softstate->dip, &tmp_dma_attr,
	                     DDI_DMA_DONTWAIT,NULL, &memchunk->dmahandle);
	if(tmp_state!= DDI_SUCCESS)
	{
		if(tmp_state==DDI_DMA_BADATTR){
			printf("ddi_dma_alloc_handle failed: bad attr\n");
		} else if(tmp_state==DDI_DMA_NORESOURCES){
			printf("ddi_dma_alloc_handle failed: no resources\n");
		} else {
			printf("ddi_dma_alloc_handle failed\n");
		}
		return (ENOMEM);
	}
	printf("allocated dmahandle=%p\n",memchunk->dmahandle);

	tmp_state=DDI_FAILURE;

	/* XXX whastupwiththis */
	tmp_state=ddi_dma_mem_alloc(memchunk->dmahandle, memsize,
					&dev_attr,
#ifdef NORMAL
					DDI_DMA_CONSISTENT,
#else
					DDI_DMA_STREAMING,
#endif
					DDI_DMA_DONTWAIT, NULL,
					&memchunk->kaddr,
					&memchunk->reallen,
					&memchunk->dma_acc_handle);
	if(tmp_state!=DDI_SUCCESS){
		printf("mem_alloc for size %d failed\n",memsize);
		return ENOMEM;
	}

	dmamem_debugprint("allocated acc_handle=%p\n",memchunk->dma_acc_handle);
	dmamem_debugprint("kaddr=%p, reallen=%ld\n",memchunk->kaddr,memchunk->reallen);


	/* Yes, this is ugly, doing seemingly redundant calls.
	 * Need to call addr_bind_handle, to get phys address.
	 * However, need to redundantly call ddi_umem_lock,
	 *   just to get a ddi_umem_cookie, so that we
	 *   can actually use devmap(9e)
	 */
#ifdef NORMAL
	tmp_state=ddi_dma_addr_bind_handle(memchunk->dmahandle,
	        NULL, memchunk->kaddr, memchunk->reallen,
		DDI_DMA_RDWR|DDI_DMA_CONSISTENT, DDI_DMA_SLEEP, NULL,
		&memchunk->dmacookie, &memchunk->ccount);
#else
	tmp_state=ddi_dma_addr_bind_handle(memchunk->dmahandle,
	        NULL, memchunk->kaddr, memchunk->reallen,
		DDI_DMA_RDWR|DDI_DMA_STREAMING, DDI_DMA_SLEEP, NULL,
		&memchunk->dmacookie, &memchunk->ccount);
#endif

	if(tmp_state!=DDI_SUCCESS){
		printf("ddi_dma_addr_bind_handle failed\n");
		free_memchunk(memchunk);
		return ENXIO;
	}
	if(memchunk->ccount>1){
		/* This should never happen, since tmp_dma_attr_t
		 * specified sglen=1
		 */
		printf("ERROR: more than one cookie returned on bind\n");
		free_memchunk(memchunk);
		return ENXIO;
	}

	/* XXX note: this is 32bit address only! */
	dstruct->physaddr=memchunk->dmacookie.dmac_address;
	
	dmamem_debugprint("Success for ddi_dma_addr_bind_handle: memsize=%d,physaddr=%x\n",
	       memsize, dstruct->physaddr);


#ifndef USE_MMAP_ONLY
	/* Agr.. I hate that we have to essentialy bind the stupid
	 * thing TWICE!!!
	 * The sole purpose of this is to get the magic cookie that
	 * we can then use in devmap
	 */

	tmp_state=ddi_umem_lock(memchunk->kaddr, memchunk->reallen,
	       DDI_UMEMLOCK_READ|DDI_UMEMLOCK_WRITE,
	       &memchunk->umem_cookie);

	if(tmp_state!=0){
		printf("ERR: ddi_umem_lock failed\n");
	}

#else /*USE_MMAP_ONLY *//* only using mmap(9e), so dont have to fuss */

	tmp_state=0;
	
#endif

	if(tmp_state==0){
		set_free_offset(softstate,memchunk);
		memchunk->memlen=memsize;
		dstruct->mmap_offset=memchunk->offset;
		dmamem_debugprint("mmap offset for new chunk is 0x%x\n",
			  dstruct->mmap_offset);
		return 0;
	}

	/* do NOT return tmp_state. return some suitable value */
	return EINVAL;

}



/* This routine can be called by dmamem_attach (if errors occur)
 * and dmamem_detach (on normal detach)
 * Idea is to clean up any mutexes, register mappings, etc. nicely.
 * We also free the instance-softstate.
 * We do NOT remove the minor node, however. Only detach does
 * that currently.
 */
static void dmamem_cleanup(struct dmamem_softstate *softstate){
	int instance;

	if(softstate==NULL){
		/* yes, paranoia is good*/
		cmn_err(CE_WARN,"ERROR: dmamem_cleanup called with NULL");
		return;
	}
   	instance=ddi_get_instance(softstate->dip);

	dmamem_freemem(softstate);

	ddi_soft_state_free(dmamem_glob_soft_handle,instance);
	
}



#ifdef USE_MMAP_ONLY
/* return pagenumber for the SINGLE PAGE, at offset 'off' */
static int dmamem_mmap(dev_t dev, off_t offset, int prot)
{
	int instance=DEVT_TO_INST(dev);
	int pagenum, relative_offset;
	struct dmamem_softstate *softstate;
	struct memchunk *memchunk;

	softstate=ddi_get_soft_state(dmamem_glob_soft_handle, instance);
	if(softstate==NULL){
		printf("mmap could not get softstate for instance %d\n",
		        instance);
		return (ENXIO);
	}
	if(offset & 0xfff){
		dmamem_debugprint("ERR: mmap offset not page-aligned\n");
		return (-1);
	}

	memchunk=find_chunk_hard_way(softstate, offset);
	if(memchunk==NULL){
		return (-1);
	}

	/* OKay, we know that the user offset lies "somewhere" in
	 * the memchunk we have now located. now figure out where
	 */
	relative_offset=offset - memchunk->offset;

	pagenum=hat_getkpfnum(memchunk->kaddr + relative_offset);
	if(pagenum== -1){
		printf("mmap: NOTE: hat_getkpfnum FAILED\n");
	}
	return pagenum;
}

#else /*USE_MMAP_ONLY*/

/* Not using mmap - using newer devmap framework.
 * we can map an entire memory range at once this way, unlike devmap_mmap
 */
static int dmamem_devmap(dev_t dev, devmap_cookie_t cookie, offset_t offset,
		      size_t fulllen, size_t *mappedlen, uint_t model)
{
	int instance=DEVT_TO_INST(dev);
	int status;
	struct dmamem_softstate *softstate;
	struct memchunk *memchunk;
	uint32_t len= fulllen& 0x7fffffff;

	softstate=ddi_get_soft_state(dmamem_glob_soft_handle, instance);
	if(softstate==NULL){
		return (ENXIO);
	}
	dmamem_debugprint("devmap called with offset %x, len %d\n",
		      offset, len);

	memchunk=find_chunk_by_offset(softstate, offset);
	if(memchunk==NULL){
		return (-1);
	}
	
	/* user must map the ENTIRE memory chunk, or we assume that
	 * the user program has gotten confused, in which case we reject it
	 */
	if(len > memchunk->memlen){
		return EINVAL;
	}
	if((offset<0)||(len<0)){
		return EINVAL;
	}

	status=devmap_umem_setup(cookie, softstate->dip, NULL,
			softstate->umem_cookie, 0, len,
			PROT_ALL, DEVMAP_DEFAULTS,&dev_attr);

	dmamem_debugprint("status of devmap_umem_setup=%d\n",status);

	*mappedlen=len;

	return(status);

}

#endif

/*
 * _init() has already been called. This routine gets called
 * once for EVERY MATCHING PIECE OF HARDWARE.
 * It gets called for initial "attach", and also for resume(/suspend)
 * You would usually allocate space for state variables, and
 * map device registers in here.
 *
 * To avoid reboot loops, be very very careful putting things in here.
 * If you want to put extra debug type stuff in, it's safer putting
 * it in dmamem_open() or someplace like that, that gets called after
 * the driver is safely loaded
 *
 * Be sure to free/deallocate anything you create here, in
 *  dmamem_detach()
 */
static int dmamem_attach(dev_info_t *dip, ddi_attach_cmd_t cmd){
   	int instance;
	struct dmamem_softstate *softstate;

	/* default debug level to **1**
	 * You can override in the .conf file with "debug-level=0".
	 * When you have a production-ready driver, change
	 * default to 0
	 */
	debug_level=ddi_prop_get_int(DDI_DEV_T_ANY,dip, DDI_PROP_DONTPASS,
                                    "debug-level", 1);

	dmamem_debugprint("dmamem: dmamem_attach called\n");

   	switch(cmd){
	 case DDI_RESUME:
	   	cmn_err(CE_NOTE, "dmamem: Ignoring attach_RESUME\n");
		return DDI_FAILURE;
	 case DDI_PM_RESUME:
		cmn_err(CE_NOTE, "dmamem: Ignoring attach_PM_RESUME\n");
		return DDI_FAILURE;
	 case DDI_ATTACH:
	   	break;
	 default:
	   	cmn_err(CE_NOTE, "dmamem: unknown attach command\n");
	   	return DDI_FAILURE;
	}
   	instance=ddi_get_instance(dip);
   	dmamem_debugprint("dmamem: attach for instance %d\n", instance);


	/* We dont have to check for being over MAX_INSTANCES, because
	 * these two calls do it implicitly
	 */
	if (ddi_soft_state_zalloc(dmamem_glob_soft_handle, instance) != DDI_SUCCESS) {
		cmn_err(CE_WARN, "dmamem: %d: Failed to alloc soft state",
		    instance);
		return (DDI_FAILURE);
	}
	softstate=ddi_get_soft_state(dmamem_glob_soft_handle,instance);
	if(softstate==NULL){
		cmn_err(CE_WARN, "dmamem: attach for instance %d: Failed to get soft state",
			instance);
		ddi_soft_state_free(dmamem_glob_soft_handle,instance);
		return (DDI_FAILURE);
	}

	softstate->dip = dip;

	/* you will probably want this at some point or another */
	mutex_init(&softstate->instance_mutex, NULL, MUTEX_DRIVER, NULL);



	/* This will create a device node in the appropriate place
	 * under /devices/pci[*]/XXXXXX
	 *
	 * To get an automatic matching link in /dev/XXXX, you need to 
	 * add your own line to /etc/devlink.tab
	 */
	 
	ddi_create_minor_node(dip, DEVNODE_NAME, S_IFCHR,
			  INST_TO_NODENUM(instance),
			  DEVNODE_TYPE, 0);


	/* Dont forget to make ddi_remove_minor_node match in detach() */

	return DDI_SUCCESS;

	/* if you add anything to this routine, make ultra-paranoid
	 * checks on return values and return (DDI_FAILURE)
	 * if ANYTHING doesnt go completely 100% okay
	 */
}


static int dmamem_open(dev_t *dev, int openflags, int otyp, cred_t *credp){
	int instance=DEVT_TO_INST(*dev);
	struct dmamem_softstate *softstate;

   	dmamem_debugprint("dmamem: open called\n");

	/* even if you dont USE this, it is a check to see if we need
	 * to return ENXIO
	 */
	softstate=ddi_get_soft_state(dmamem_glob_soft_handle, instance);
	if(softstate==NULL){
		/* This is technically NOT an error; it is expected.
		 * We return ENXIO to support "deferred attach"
		 */
		return (ENXIO);
	}

	if(softstate->openflag!=0){
		printf("ERROR: already open\n");
		return EBUSY;
	}
	softstate->openflag=1;

	/* return 0 on okay, or an errno otherwise */
	return 0;
}



static int dmamem_read(dev_t dev, struct uio* uiop, cred_t * credp){
	/*int instance=DEVT_TO_INST(dev);*/
	
   	dmamem_debugprint("dmamem: read called\n");
	/* If you are handling things directly here rather than through
	 * physio, make sure that uiop->resid is set to 0 on good read
	 */
	return (ENXIO);
}

static int dmamem_write(dev_t dev, struct uio *uiop, cred_t *credp){
	/*int instance=DEVT_TO_INST(dev);*/

   	dmamem_debugprint("dmamem: write called\n");
	/* If you are handling things directly here rather than through
	 * physio, make sure that uiop->resid is set to 0 on good write
	 */
   	return (ENXIO);
}

static int dmamem_close(dev_t dev, int openflags, int otyp, cred_t *credp){
	int instance=DEVT_TO_INST(dev);
	struct dmamem_softstate *softstate;

   	dmamem_debugprint("dmamem: close called\n");

	softstate=ddi_get_soft_state(dmamem_glob_soft_handle, instance);
	if(softstate==NULL){
		printf("ERROR: close() could not get softstate??\n");
		return (ENXIO);
	}

	if(softstate->openflag==0){
		printf("ERROR: already closed\n");
		return ENXIO;
	}
	
	dmamem_freemem(softstate);
	softstate->openflag=0;

   	return (0);
}

static int dmamem_ioctl(dev_t dev, int cmd, intptr_t arg, int flags,
                      cred_t *credp, int *rvalp){
	struct dmamem meminfo;
	int instance=DEVT_TO_INST(dev);
	struct dmamem_softstate *softstate;
	void*argp=(void*)arg;
	int retval=0;

	softstate=ddi_get_soft_state(dmamem_glob_soft_handle, instance);
	if(softstate==NULL){
		cmn_err(CE_NOTE,"dmamem_ioctl failed: cannot get softstate?");
		return DDI_FAILURE;
	}

   	dmamem_debugprint("dmamem: ioctl called\n");

	switch(cmd){
	    case DMAMEM_IOCTL_VERSIONSTR:
		if(ddi_copyout(DMAMEM_VERSIONSTR, argp,
			       strlen(DMAMEM_VERSIONSTR)+1, flags))
		{
			printf("Copyout failed\n");
			retval=ENXIO;
		}
		break;

	    case DMAMEM_IOCTL_ALLOCATE:
		if(ddi_copyin(argp, &meminfo, sizeof(meminfo), flags)){
			printf("Copyin failed??\n");
			return ENXIO;
		}
		retval=create_dmamem(softstate, &meminfo);
		if(ddi_copyout(&meminfo, argp, sizeof(meminfo), flags)){
			printf("Copyout failed??\n");
			return ENXIO;
		}
		break;

	    default:
		dmamem_debugprint("Invalid ioctl attempted\n");
		return (EINVAL);
	}

   	return (retval);
}

static int dmamem_detach(dev_info_t *dip, ddi_detach_cmd_t cmd){
   	int instance;
	struct dmamem_softstate *softstate;
   	instance=ddi_get_instance(dip);

	dmamem_debugprint("dmamem: detach for instance %d\n", instance);

	softstate=ddi_get_soft_state(dmamem_glob_soft_handle, instance);
	if(softstate==NULL){
		cmn_err(CE_NOTE,"dmamem_detach failed: cannot get softstate?");
		return DDI_FAILURE;
	}

	dmamem_cleanup(softstate);

	ddi_remove_minor_node(dip, DEVNODE_NAME);

	return (DDI_SUCCESS);
		     
}

