#include <sys/types.h>
#include <sys/ioccom.h>

#define DMAMEM_DEVICE "/dev/dmamem"


#define DMAMEM_IOCTL_BASE		'G'
#define DMAMEM_IOCTL_VERSIONSTR	_IOR(DMAMEM_IOCTL_BASE,  0, char )
#define DMAMEM_IOCTL_ALLOCATE	_IOWR(DMAMEM_IOCTL_BASE,  8, struct dmamem)

/*
 * Call DMAMEM_IOCTL_VERSIONSTR with a pointer to char[DMAMEM_VERSIONSTRLEN]
 * You can then strncmp or printf the results.
 * The resulting string should be equal to DMAMEM_VERSIONSTR, otherwise
 * you have an incompatible version of the driver.
 * 
 */
#define DMAMEM_VERSIONSTRLEN 20
#define DMAMEM_VERSIONSTR "dmamem version 1.0"



/* This structure gets used for both allocate, and deallocate.
 * To allocate, just set the memsize you want for a single physical chunk.
 * If it returns fail, feel free to call again with a smaller size.
 *
 * To deallocate, ignore everything except setting the mmap_offset.
 *
 * Technically, you shouldnt call deallocate, until the
 * driver can handle the devmap entry-point, because solaris handling of
 * mmap(9e) gets ugly once you start calling munmap, unfortunately.
 * Just assume that all the resources will get cleaned up when your
 * program exits.
 * But I document the aspired-to calling sequence for future reference.
 *
 */
struct dmamem {
	uint32_t memsize;     /* caller sets this and nothing else */
	off_t    mmap_offset; /* pass result to mmap, to map the memchunk */
	uint32_t physaddr;    /* only need this for programming hardware */
};


/* This defines the maximum number of times you can expect
 * to call DMAMEM_IOCTL_ALLOCATE
 * Note that this is a _maximum_. There is no guarantee it will
 *  _succeed_ this many times, with the size of memory chunks you want.
 */
#define DMAMEM_MAX_CHUNKS 20

