/* Simple test prog to see if we can mmap memory, from /dev/dmamem
 * Will mmap, tweak some bits, and then attempt to show the stuff
 * in physical memory.
 *
 * Note that this must be run as root, since it needs to open up /dev/mem
 * to verify that we successfully mapped the physical mem we think we did.
 */

#include <stdio.h>
#include <unistd.h>

#include <string.h> /* for memset */
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>

#include <time.h>


#include "dmamem.h"

#define MAPSIZE 65536

int validate_mem(int devfd, struct dmamem *meminfo)
{
	char *mapaddr, *rawmem;
	char string1[100];
	char string2[100];
	int fd2;
	int offset=meminfo->mmap_offset;

	printf("validating mem allegedly at physaddr 0x%x, m_offset 0x%x\n",
	       meminfo->physaddr, offset);

	mapaddr=mmap((caddr_t)0, MAPSIZE, (PROT_READ|PROT_WRITE),
	             MAP_SHARED, devfd, offset);
	if(mapaddr == MAP_FAILED){
		perror("could not mmap memory");
		exit(1);
	}
	printf("mmap of DMA mem returned %p\n",mapaddr);

	puts("Successfully did mmap of DMA mem.");

	fd2=open("/dev/mem",O_RDONLY);
	if(fd2<0){
		perror("could not open /dev/mem");
		exit(errno);
	}
	rawmem=mmap((caddr_t)0, MAPSIZE, (PROT_READ),
	             MAP_SHARED, fd2, meminfo->physaddr);
	if(rawmem == MAP_FAILED){
		perror("could not mmap raw mem");
		exit(1);
	}
	printf("mmap of raw addr %p mapped to %x\n",rawmem,meminfo->physaddr);


	memset(mapaddr, 0, 40); /* zero out the page a little first */

	sprintf(string1, "The time is %ld\n",time(NULL));
	memcpy(mapaddr, string1, strlen(string1));

	printf("Looking for:  %s\n",string1);

	memcpy(string2, rawmem, strlen(string1)+1);
	string2[strlen(string1)]='\0';

	printf("in raw memory, we see %s\n", string2);
	if(strcmp(string1,string2)==0){
		puts("MATCH!");
	} else {
		puts("No match.");
	}

	close(fd2);

	return 0;
}

int main(int argc, char *argv[]){
	char *devpath=DMAMEM_DEVICE;
	char string1[100];
	int fd;

	struct dmamem memreq1 = {
		MAPSIZE, 0, 0
	};
	struct dmamem memreq2 = {
		MAPSIZE, 0, 0
	};

	fd=open(devpath, O_RDWR);
	if(fd<0){
		perror("could not open file");
		exit(1);
	}
	if(ioctl(fd, DMAMEM_IOCTL_VERSIONSTR,string1)!=0){
		perror(" failed to get dmamem version string");
		exit(1);
	}
	printf("version string is: %s\n",string1);
	if(strncmp(string1, DMAMEM_VERSIONSTR,DMAMEM_VERSIONSTRLEN) !=0){
		puts("Mismatch of version string. Cannot continue");
		exit(1);
	}
	puts("version string matches! ");

	if(ioctl(fd, DMAMEM_IOCTL_ALLOCATE,&memreq1)!=0){
		perror("alloc failed");
		exit(errno);
	} else {
		printf("ioctl alloc succeded: physaddr=%x\n",memreq1.physaddr);
	}
	if(ioctl(fd, DMAMEM_IOCTL_ALLOCATE,&memreq2)!=0){
		perror("alloc failed");
		exit(errno);
	} else {
		printf("ioctlalloc succeded: physaddr=%x\n",memreq2.physaddr);
	}

	puts("validating first alloc");
	validate_mem(fd, &memreq1);

	puts("validating second alloc");
	validate_mem(fd, &memreq2);

	/* You MUST hold this open for as long as you want the allocated
	 * memory to stay allocated. Only close it at end of program
	 */
	close(fd);

	exit(0);
}
